from __future__ import annotations

import sys
from typing import TYPE_CHECKING, cast

import anyio
from advanced_alchemy.exceptions import ConflictError, NotFoundError
from click import echo, group, option, prompt
from litestar.cli._utils import LitestarGroup

from litestar_users.utils import async_session, get_litestar_users_plugin, get_user_service

if TYPE_CHECKING:
    from litestar import Litestar


@group(cls=LitestarGroup, name="users")
def user_management_group() -> None:
    """Manage users."""


@user_management_group.command(
    name="create-user",
    help="Create a new user in the database.",
    context_settings={"ignore_unknown_options": True, "allow_extra_args": True},
)
@option("--email", help="The user's email address.")
@option("--password", help="The new user's login password.")
@option("--is-active", is_flag=True, default=False, type=bool, help="Set the user as active.")
@option("--is-verified", is_flag=True, default=False, type=bool, help="Set the user as verified.")
@option("--id", "user_id", help="Set the user ID.")
@option(
    "--bool-attrs", "-b", "booleans", multiple=True, help="Set one or more custom boolean attribute key-value pairs."
)
@option("--float-attrs", "-f", "floats", multiple=True, help="Set one or more custom float attribute key-value pairs.")
@option(
    "--int-attrs", "-i", "integers", multiple=True, help="Set one or more custom integer attribute key-value pairs."
)
@option("--str-attrs", "-s", "strings", multiple=True, help="Set one or more custom string attribute key-value pairs.")
def create_user(
    app: Litestar,
    email: str,
    password: str,
    is_active: bool,
    is_verified: bool,
    user_id: str | None,
    booleans: tuple[str],
    integers: tuple[str],
    floats: tuple[str],
    strings: tuple[str],
) -> None:
    """Create a new user in the database."""
    kwargs: dict[str, str | int | float | bool] = {}
    try:
        kwargs.update(
            {
                key: value.lower() in ("1", "true", "t", "yes", "y")
                for arg in booleans
                for key, value in [arg.split("=")]
            }
        )
        kwargs.update({key: int(value) for arg in integers for key, value in [arg.split("=")]})
        kwargs.update({key: float(value) for arg in floats for key, value in [arg.split("=")]})
        kwargs.update({key: value for arg in strings for key, value in [arg.split("=")]})
    except ValueError as e:
        echo(f"Error: {e}", err=True)
        sys.exit(1)
    if user_id:
        kwargs["id"] = int(user_id) if user_id.isnumeric() else user_id

    litestar_users_config = get_litestar_users_plugin(app)._config

    email = email or prompt("User email")
    password = password or prompt("User password", hide_input=True, confirmation_prompt=True)

    async def _create_user() -> None:
        async with async_session(app) as session:
            user_service = get_user_service(app, session)
            password_hash = user_service.password_manager.hash(password)
            try:
                user = await user_service.add_user(
                    user=litestar_users_config.user_model(email=email, password_hash=password_hash, **kwargs),
                    activate=is_active,
                    verify=is_verified,
                )
                await session.commit()
                echo(f"User {user.id} created successfully.")
            except ConflictError as e:
                # could be caught IntegrityError or unique collision
                msg = e.__cause__ if e.__cause__ else e
                echo(f"Error: {msg}", err=True)
                sys.exit(1)
            except TypeError as e:
                echo(f"Error: {e}", err=True)
                sys.exit(1)

    anyio.run(_create_user)


@user_management_group.command(name="create-role", help="Create a new role in the database.")
@option("--name")
@option("--description")
def create_role(app: Litestar, name: str | None, description: str | None) -> None:
    """Create a new role in the database."""

    litestar_users_config = get_litestar_users_plugin(app)._config

    name = name or prompt("Role name")

    async def _create_role() -> None:
        async with async_session(app) as session:
            user_service = get_user_service(app, session)
            if litestar_users_config.role_model is None:
                echo("Role model is not defined")
                sys.exit(1)
            try:
                role = await user_service.add_role(litestar_users_config.role_model(name=name, description=description))
                await session.commit()
                echo(f"Role {role.id} created successfully.")
            except ConflictError as e:
                # could be caught IntegrityError or unique collision
                msg = e.__cause__ if e.__cause__ else e
                echo(f"Error: {msg}", err=True)
                sys.exit(1)

    anyio.run(_create_role)


@user_management_group.command(name="assign-role", help="Assign a role to a user.")
@option("--email")
@option("--role")
def assign_role(
    app: Litestar,
    email: str | None,
    role: str | None,
) -> None:
    """Assign a role to a user."""

    litestar_users_config = get_litestar_users_plugin(app)._config
    if litestar_users_config.role_model is None:
        echo("Role model is not defined")
        sys.exit(1)

    email = email or prompt("User email")
    role = role or cast(str, prompt("Role", type=str))

    async def _assign_role() -> None:
        async with async_session(app) as session:
            user_service = get_user_service(app, session)
            try:
                user = await user_service.get_user_by(email=email)
                if user is None:
                    raise NotFoundError()
            except NotFoundError:
                echo("User not found", err=True)
                sys.exit(1)
            try:
                role_db = await user_service.get_role_by_name(role)  # type: ignore[arg-type]
            except NotFoundError:
                echo("Role not found", err=True)
                sys.exit(1)
            await user_service.assign_role(user.id, role_db.id)
        echo(f"Role {role} assigned to user {email} successfully.")

    anyio.run(_assign_role)



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, AsyncIterator

from advanced_alchemy.extensions.litestar.plugins import SQLAlchemyInitPlugin
from litestar.exceptions import ImproperlyConfiguredException
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker

from litestar_users.adapter.sqlalchemy.repository import SQLAlchemyRoleRepository

if TYPE_CHECKING:
    from litestar import Litestar
    from sqlalchemy.ext.asyncio import AsyncSession

    from litestar_users import LitestarUsersPlugin
    from litestar_users.service import BaseUserService


def get_litestar_users_plugin(app: Litestar) -> LitestarUsersPlugin:
    """Get the LitestarUsersPlugin from the Litestar application."""
    from litestar_users import LitestarUsersPlugin

    try:
        return app.plugins.get(LitestarUsersPlugin)
    except KeyError as e:
        raise ImproperlyConfiguredException("The LitestarUsersPlugin is missing from the application") from e


def get_sqlalchemy_plugin(app: Litestar) -> SQLAlchemyInitPlugin:
    """Get the SQLAlchemyPlugin from the Litestar application."""
    try:
        return app.plugins.get(SQLAlchemyInitPlugin)
    except KeyError as e:
        raise ImproperlyConfiguredException("The SQLAlchemyPlugin is missing from the application") from e


@asynccontextmanager
async def async_session(app: Litestar) -> AsyncIterator[AsyncSession]:
    """Get an async session outside of a Litestar request context."""

    sqlalchemy_config = get_sqlalchemy_plugin(app)._config
    engine = sqlalchemy_config.get_engine()
    if not isinstance(engine, AsyncEngine):
        raise ImproperlyConfiguredException("The SQLAlchemy engine must be async")
    sessionmaker = async_sessionmaker(bind=engine, expire_on_commit=False)

    async with sessionmaker() as session:
        yield session


def get_user_service(app: Litestar, session: AsyncSession) -> BaseUserService[Any, Any]:
    """Get a `UserService` instance outside of a Litestar request context."""

    config = get_litestar_users_plugin(app)._config

    user_repository = config.user_repository_class(session, config.user_model)
    role_repository: SQLAlchemyRoleRepository | None = (
        SQLAlchemyRoleRepository(session, config.role_model) if config.role_model else None
    )
    return config.user_service_class(
        user_auth_identifier=config.user_auth_identifier,
        user_repository=user_repository,
        role_repository=role_repository,
        secret=config.secret,
        hash_schemes=config.hash_schemes,
    )



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar

from litestar.repository.abc import AbstractAsyncRepository

__all__ = ["AbstractRoleRepository", "AbstractUserRepository"]


T = TypeVar("T")
R = TypeVar("R")
URepoT = TypeVar("URepoT", bound="AbstractUserRepository")


class AbstractUserRepository(AbstractAsyncRepository[T], Generic[T]):
    """Interface for user persistence interaction."""

    @abstractmethod
    async def _update(self, user: T, data: dict[str, Any]) -> T:
        """Update a user that is already present in the db session."""


class AbstractRoleRepository(AbstractAsyncRepository[R], Generic[R, T]):
    """Interface for role persistence interaction."""

    @abstractmethod
    async def assign_role(self, user: T, role: R) -> T:
        """Add a role to a user.

        Args:
            user: The user to receive the role.
            role: The role to add to the user.
        """

    @abstractmethod
    async def revoke_role(self, user: T, role: R) -> T:
        """Revoke a role from a user.

        Args:
            user: The user to revoke the role from.
            role: The role to revoke from the user.
        """



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Generic

from advanced_alchemy.repository import SQLAlchemyAsyncRepository
from litestar.exceptions import ImproperlyConfiguredException

from litestar_users.adapter.sqlalchemy.protocols import SQLARoleT, SQLAUserT

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession

__all__ = ["SQLAlchemyRoleRepository", "SQLAlchemyUserRepository"]


class SQLAlchemyUserRepository(SQLAlchemyAsyncRepository[SQLAUserT], Generic[SQLAUserT]):
    """SQLAlchemy implementation of user persistence layer."""

    def __init__(self, session: AsyncSession, model_type: type[SQLAUserT]) -> None:
        """Repository for users.

        Args:
            session: Session managing the unit-of-work for the operation.
            model_type: A subclass of `SQLAlchemyUserModel`.
        """
        self.model_type = model_type
        super().__init__(session=session)

    async def _update(self, user: SQLAUserT, data: dict[str, Any]) -> SQLAUserT:
        for key, value in data.items():
            setattr(user, key, value)

        await self.session.commit()
        return user


class SQLAlchemyRoleRepository(SQLAlchemyAsyncRepository[SQLARoleT], Generic[SQLARoleT, SQLAUserT]):
    """SQLAlchemy implementation of role persistence layer."""

    def __init__(self, session: AsyncSession, model_type: type[SQLARoleT]) -> None:
        """Repository for users.

        Args:
            session: Session managing the unit-of-work for the operation.
            model_type: A subclass of `SQLAlchemyRoleModel`.
        """
        self.model_type = model_type
        super().__init__(session=session)

    async def assign_role(self, user: SQLAUserT, role: SQLARoleT) -> SQLAUserT:
        """Add a role to a user.

        Args:
            user: The user to receive the role.
            role: The role to add to the user.
        """
        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("User.roles is not set")
        user.roles.append(role)  # pyright: ignore
        await self.session.commit()
        return user

    async def revoke_role(self, user: SQLAUserT, role: SQLARoleT) -> SQLAUserT:
        """Revoke a role from a user.

        Args:
            user: The user to revoke the role from.
            role: The role to revoke from the user.
        """
        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("User.roles is not set")
        user.roles.remove(role)  # pyright: ignore
        await self.session.commit()
        return user



from __future__ import annotations

from typing import TYPE_CHECKING

from advanced_alchemy.exceptions import ConflictError, NotFoundError, RepositoryError
from litestar.exceptions import (
    HTTPException,
    InternalServerException,
    NotFoundException,
)
from litestar.middleware.exceptions._debug_response import create_debug_response
from litestar.middleware.exceptions.middleware import create_exception_response

__all__ = [
    "ExpiredTokenException",
    "InvalidException",
    "InvalidTokenException",
    "TokenException",
    "token_exception_handler",
]


if TYPE_CHECKING:
    from litestar import Request, Response


class TokenException(Exception):
    """Base token exception."""


class InvalidTokenException(TokenException):
    """Raise when a JWT is found to be invalid."""


class ExpiredTokenException(TokenException):
    """Raise when a JWT is found to be expired."""


class InvalidException(HTTPException):
    """Generic HTTP invalid exception."""

    status_code = 400


class ConflictException(HTTPException):
    status_code = 409


def token_exception_handler(request: Request, exception: TokenException) -> Response:
    """Transform token exceptions to HTTP exceptions."""
    http_exception: type[HTTPException]
    if isinstance(exception, (InvalidTokenException, ExpiredTokenException)):
        http_exception = InvalidException
    else:
        http_exception = InternalServerException
    if request.app.debug:
        return create_debug_response(request, exception)
    return create_exception_response(request=request, exc=http_exception(detail=str(exception)))


def repository_exception_to_http_response(request: Request, exception: RepositoryError) -> Response:
    """Transform repository exceptions to HTTP exceptions.

    Args:
        request: The request that experienced the exception.
        exception: Exception raised during handling of the request.

    Returns:
        Exception response appropriate to the type of original exception.
    """
    http_exception: type[HTTPException]
    if isinstance(exception, NotFoundError):
        http_exception = NotFoundException
    elif isinstance(exception, ConflictError):
        http_exception = ConflictException
    else:
        http_exception = InternalServerException
    if request.app.debug and not request.app.state.get("testing"):
        return create_debug_response(request, exception)
    return create_exception_response(request=request, exc=http_exception(detail=str(exception)))



from __future__ import annotations

from dataclasses import dataclass, field, is_dataclass
from typing import TYPE_CHECKING, Any, Generic

from litestar.exceptions import ImproperlyConfiguredException
from litestar.security.session_auth import SessionAuth

from litestar_users.adapter.sqlalchemy.repository import SQLAlchemyUserRepository
from litestar_users.protocols import RoleT, UserT
from litestar_users.schema import AuthenticationSchema

__all__ = [
    "AuthHandlerConfig",
    "CurrentUserHandlerConfig",
    "PasswordResetHandlerConfig",
    "RegisterHandlerConfig",
    "RoleManagementHandlerConfig",
    "LitestarUsersConfig",
    "UserManagementHandlerConfig",
    "VerificationHandlerConfig",
]

if TYPE_CHECKING:
    from advanced_alchemy.extensions.litestar.dto import SQLAlchemyDTO
    from litestar.contrib.jwt import JWTAuth, JWTCookieAuth
    from litestar.contrib.pydantic import PydanticDTO
    from litestar.dto import DataclassDTO, MsgspecDTO
    from litestar.middleware.session.base import BaseBackendConfig
    from litestar.types import Guard

    from litestar_users.service import BaseUserService

USER_CREATE_DTO_EXCLUDED_FIELDS = {"password_hash"}
USER_READ_DTO_EXCLUDED_FIELDS = {"password"}
DEFAULT_USER_AUTH_IDENTIFIER = "email"


@dataclass
class AuthHandlerConfig:
    """Configuration for user authentication route handlers.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    login_path: str = "/login"
    """The path for the user authentication/login route."""
    logout_path: str = "/logout"
    """The path for the logout route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class CurrentUserHandlerConfig:
    """Configuration for the current-user route handler.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/users/me"
    """The path to get or update the currently logged-in user."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class PasswordResetHandlerConfig:
    """Configuration for the forgot-password and reset-password route handlers.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    forgot_path: str = "/forgot-password"
    """The path for the forgot-password route."""
    reset_path: str = "/reset-password"
    """The path for the reset-password route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class RegisterHandlerConfig:
    """Configuration for the user registration route handler.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/register"
    """The path for the registration/signup route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class RoleManagementHandlerConfig:
    """Configuration for the role management route handlers.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    path_prefix: str = "/users/roles"
    """The prefix for the router path."""
    assign_role_path: str = "/assign"
    """The path for the role assignment router."""
    revoke_role_path: str = "/revoke"
    """The path for the role revokement router."""
    guards: list[Guard] = field(default_factory=list)
    """A list of callable [Guards][litestar.types.Guard] that determines who is authorized to manage roles."""
    opt: dict[str, Any] = field(default_factory=dict)
    """Optional route handler [opts][litestar.controller.Controller.opt] to provide additional context to Guards."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class UserManagementHandlerConfig:
    """Configuration for user management (read, update, delete) route handlers.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.

    Note:
    - These routes make use of Litestar `Guard`s to require authorization. Callers require admin or similar privileges.
    """

    path_prefix: str = "/users"
    """The prefix for the router path.

    By default, the path will be suffixed with `'/{user_id:<type>}'`.
    """
    guards: list[Guard] = field(default_factory=list)
    """A list of callable [Guards][litestar.types.Guard] that determines who is authorized to manage other users."""
    opt: dict[str, Any] = field(default_factory=dict)
    """Optional route handler [opts][litestar.controller.Controller.opt] to provide additional context to Guards."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class VerificationHandlerConfig:
    """Configuration for the user verification route handler.

    Passing an instance to `LitestarUsersConfig` will automatically take care of handler registration on the app.
    """

    path: str = "/verify"
    """The path for the verification route."""
    tags: list[str] | None = None
    """A list of string tags to append to the schema of the route handler(s)."""


@dataclass
class LitestarUsersConfig(Generic[UserT, RoleT]):
    """Configuration class for LitestarUsers."""

    auth_backend_class: type[JWTAuth | JWTCookieAuth | SessionAuth]
    """The authentication backend to use by Litestar."""
    secret: str
    """Secret string for securely signing tokens."""
    user_model: type[UserT]
    """A subclass of a `User` ORM model."""
    user_service_class: type[BaseUserService]
    """A subclass of [BaseUserService][litestar_users.service.BaseUserService]."""
    user_registration_dto: type[DataclassDTO | MsgspecDTO | PydanticDTO]
    """DTO class user for user registration."""
    user_read_dto: type[SQLAlchemyDTO]
    """A `User` model based SQLAlchemy DTO class."""
    user_update_dto: type[SQLAlchemyDTO]
    """A `User` model based SQLAlchemy DTO class."""
    user_repository_class: type[SQLAlchemyUserRepository] = SQLAlchemyUserRepository
    """The user repository class to use."""
    authentication_request_schema: Any = AuthenticationSchema
    """The schema to use for authentication requests.

    This can be a dataclass, pydantic `BaseModel` or msgspec `Struct`.
    Requires an attribute with the same name as `user_auth_identifier` as well as a `password` attribute.

    Notes:
        - Required if `user_auth_identifier` is set to a non-default value.
    """
    auth_exclude_paths: list[str] = field(default_factory=lambda: ["/schema"])
    """Paths to be excluded from authentication checks."""
    hash_schemes: list[str] = field(default_factory=lambda: ["argon2"])
    """Schemes to use for password encryption.

    Defaults to `["argon2"]`
    """
    session_backend_config: BaseBackendConfig | None = None
    """Optional backend configuration for session based authentication.

    Notes:
        - Required if `auth_backend_class` is `SessionAuth`.
    """
    role_model: type[RoleT] | None = None
    """A `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_create_dto: type[SQLAlchemyDTO] | None = None
    """A `SQLAlchemyDTO` based on a `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_read_dto: type[SQLAlchemyDTO] | None = None
    """A `SQLAlchemyDTO` based on a `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    role_update_dto: type[SQLAlchemyDTO] | None = None
    """A `SQLAlchemyDTO` based on a `Role` ORM model.

    Notes:
        - Required if `role_management_handler_config` is set.
    """
    user_auth_identifier: str = DEFAULT_USER_AUTH_IDENTIFIER
    """The identifying attribute to use during user authentication. Defaults to `'email'`.

    Changing this value requires setting `authentication_request_schema` as well, which would allow login via e.g. `username` instead.

    Notes:
        - The attribute must be present on the `User` database model and must have a unique value.
    """
    auth_handler_config: AuthHandlerConfig | None = None
    """Optional instance of [AuthHandlerConfig][litestar_users.config.AuthHandlerConfig]. If set, registers the route
    handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    current_user_handler_config: CurrentUserHandlerConfig | None = None
    """Optional current-user route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    password_reset_handler_config: PasswordResetHandlerConfig | None = None
    """Optional password reset route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    register_handler_config: RegisterHandlerConfig | None = None
    """Optional registration/signup route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    role_management_handler_config: RoleManagementHandlerConfig | None = None
    """Optional role management route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    user_management_handler_config: UserManagementHandlerConfig | None = None
    """Optional user management route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """
    verification_handler_config: VerificationHandlerConfig | None = None
    """Optional user verification route handler configuration. If set, registers the route handler(s) on the app.

    Notes:
        - At least one route handler config must be set.
    """

    def __post_init__(self) -> None:
        """Validate the configuration.

        - A session backend must be configured if `auth_backend_class` is `SessionAuth`.
        - At least one route handler must be configured.
        - `role_model`, `role_create_dto`, `role_read_dto` and `role_update_dto` are required fields if
            `role_management_handler_config` is configured.
        """
        if self.auth_backend_class == SessionAuth and not self.session_backend_config:
            raise ImproperlyConfiguredException(
                'session_backend_config must be set when auth_backend is set to "session"'
            )
        handler_configs = [
            "auth_handler_config",
            "current_user_handler_config",
            "password_reset_handler_config",
            "register_handler_config",
            "role_management_handler_config",
            "user_management_handler_config",
            "verification_handler_config",
        ]
        if len(self.secret) not in [16, 24, 32]:
            raise ImproperlyConfiguredException("secret must be 16, 24 or 32 characters")
        if all(getattr(self, config) is None for config in handler_configs):
            raise ImproperlyConfiguredException("at least one route handler must be configured")
        if self.role_management_handler_config and self.role_model is None:
            raise ImproperlyConfiguredException("role_model must be set when role_management_handler_config is set")

        for field_ in self.user_read_dto.generate_field_definitions(self.user_read_dto.model_type):  # pyright: ignore
            if field_.name in USER_READ_DTO_EXCLUDED_FIELDS:
                raise ImproperlyConfiguredException(
                    f"user_read_dto fields must exclude {USER_READ_DTO_EXCLUDED_FIELDS}"
                )
        if not self.user_update_dto.config.partial:
            raise ImproperlyConfiguredException("user_update_dto.config must be partial")

        if (
            is_dataclass(self.authentication_request_schema)
            and self.user_auth_identifier not in self.authentication_request_schema.__dataclass_fields__
        ):
            raise ImproperlyConfiguredException(
                f"authentication schema class {self.authentication_request_schema} "
                f"is missing field '{self.user_auth_identifier}'"
            )
        if not is_dataclass(self.authentication_request_schema) and not hasattr(
            self.authentication_request_schema, self.user_auth_identifier
        ):
            raise ImproperlyConfiguredException(
                f"authentication schema class {self.authentication_request_schema} "
                f"is missing field '{self.user_auth_identifier}'"
            )
        # ensure password is mapped correctly
        self.user_update_dto.config.rename_fields.update({"password_hash": "password"})

        for field_ in self.user_registration_dto.generate_field_definitions(self.user_registration_dto.model_type):  # type: ignore[misc]
            if field_.name in USER_CREATE_DTO_EXCLUDED_FIELDS:
                raise ImproperlyConfiguredException(
                    f"user_registration_dto fields must exclude {USER_CREATE_DTO_EXCLUDED_FIELDS}"
                )



from __future__ import annotations

from typing import TypeVar

from sqlalchemy.orm import Mapped, declarative_mixin, mapped_column
from sqlalchemy.sql.sqltypes import Boolean, String

__all__ = [
    "SQLAlchemyRoleMixin",
    "SQLAlchemyUserMixin",
    "UserModelType",
    "RoleModelType",
]


@declarative_mixin
class SQLAlchemyUserMixin:
    """Base SQLAlchemy user mixin."""

    email: Mapped[str] = mapped_column(String(320), nullable=False, unique=True)
    password_hash: Mapped[str] = mapped_column(String(1024))
    is_active: Mapped[bool] = mapped_column(Boolean(), nullable=False, default=False)
    is_verified: Mapped[bool] = mapped_column(Boolean(), nullable=False, default=False)


@declarative_mixin
class SQLAlchemyRoleMixin:
    """Base SQLAlchemy role mixin."""

    name: Mapped[str] = mapped_column(String(255), nullable=False, unique=True)
    description: Mapped[str] = mapped_column(String(255), nullable=True)


UserModelType = TypeVar("UserModelType", bound=SQLAlchemyUserMixin)
RoleModelType = TypeVar("RoleModelType", bound=SQLAlchemyRoleMixin)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypeVar, runtime_checkable

if TYPE_CHECKING:
    from uuid import UUID


__all__ = ["RoleModelProtocol", "UserModelProtocol", "UserRegisterT"]

UserT = TypeVar("UserT", bound="UserModelProtocol")
RoleT = TypeVar("RoleT", bound="RoleModelProtocol")
UserRegisterT = TypeVar("UserRegisterT", bound="UserRegistrationProtocol")


@runtime_checkable
class RoleModelProtocol(Protocol):
    """The base role type."""

    id: UUID | int
    name: str
    description: str

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...


@runtime_checkable
class UserModelProtocol(Protocol):
    """The base user type."""

    id: UUID | int
    email: str
    password_hash: str
    is_active: bool
    is_verified: bool
    roles: list

    def __init__(self, *args: Any, **kwargs: Any) -> None: ...


@runtime_checkable
class UserRegistrationProtocol(Protocol):
    """The minimum fields required on user registration/creation."""

    email: str
    password: str



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypeVar, runtime_checkable

from advanced_alchemy.base import ModelProtocol

from sqlalchemy.orm import Mapped, MappedClassProtocol

if TYPE_CHECKING:
    from uuid import UUID

__all__ = ["SQLAlchemyRoleProtocol", "SQLAlchemyUserProtocol"]


@runtime_checkable
class SQLAlchemyRoleProtocol(ModelProtocol, MappedClassProtocol, Protocol):  # pyright: ignore
    """The base SQLAlchemy role type."""

    id: Mapped[UUID] | Mapped[int]
    name: Mapped[str]
    description: Mapped[str]


@runtime_checkable
class SQLAlchemyUserProtocol(ModelProtocol, MappedClassProtocol, Protocol):  # pyright: ignore
    """The base SQLAlchemy user type."""

    id: Mapped[UUID] | Mapped[int]
    email: Mapped[str]
    password_hash: Mapped[str]
    is_active: Mapped[bool]
    is_verified: Mapped[bool]

    def __init__(*args: Any, **kwargs: Any) -> None: ...


@runtime_checkable
class SQLAlchemyUserRoleProtocol(SQLAlchemyUserProtocol, Protocol):  # pyright: ignore
    """The base SQLAlchemy user type."""

    roles: Mapped[list[SQLAlchemyRoleProtocol]]


SQLARoleT = TypeVar("SQLARoleT", bound="SQLAlchemyRoleProtocol")
SQLAUserT = TypeVar("SQLAUserT", bound="SQLAlchemyUserProtocol | SQLAlchemyUserRoleProtocol")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union, cast

from litestar import (
    Request,
    Response,
    Router,
    delete,
    get,
    patch,
    post,
    put,
)
from litestar.contrib.jwt import JWTAuth, JWTCookieAuth
from litestar.di import Provide
from litestar.exceptions import ImproperlyConfiguredException, NotAuthorizedException, PermissionDeniedException
from litestar.security.session_auth.auth import SessionAuth

from litestar_users.adapter.sqlalchemy.protocols import SQLARoleT, SQLAUserT
from litestar_users.dependencies import provide_user_service

__all__ = [
    "get_auth_handler",
    "get_current_user_handler",
    "get_password_reset_handler",
    "get_registration_handler",
    "get_role_management_handler",
    "get_user_management_handler",
    "get_verification_handler",
]


if TYPE_CHECKING:
    from uuid import UUID

    from advanced_alchemy.extensions.litestar.dto import SQLAlchemyDTO
    from litestar.contrib.pydantic import PydanticDTO
    from litestar.dto import DataclassDTO, DTOData, MsgspecDTO
    from litestar.handlers import HTTPRouteHandler
    from litestar.types import Guard

    from litestar_users.protocols import UserRegisterT
    from litestar_users.schema import (
        ForgotPasswordSchema,
        ResetPasswordSchema,
        UserRoleSchema,
    )
    from litestar_users.service import UserServiceType


def get_registration_handler(
    path: str,
    user_registration_dto: type[DataclassDTO | MsgspecDTO | PydanticDTO],
    user_read_dto: type[SQLAlchemyDTO],
    tags: list[str] | None = None,
) -> HTTPRouteHandler:
    """Get registration route handlers.

    Args:
        path: The path for the router.
        user_registration_dto: A subclass of [UserCreateDTO][litestar_users.schema.UserCreateDTO]
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        tags: A list of string tags to append to the schema of the route handler.
    """

    @post(
        path,
        dto=user_registration_dto,
        return_dto=user_read_dto,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def register(data: DTOData[UserRegisterT], service: UserServiceType, request: Request) -> SQLAUserT:
        """Register a new user."""
        return cast(SQLAUserT, await service.register(data.as_builtins(), request))

    return register


def get_verification_handler(
    path: str,
    user_read_dto: type[SQLAlchemyDTO],
    tags: list[str] | None = None,
) -> HTTPRouteHandler:
    """Get verification route handlers.

    Args:
        path: The path for the router.
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        tags: A list of string tags to append to the schema of the route handler.
    """

    @post(
        path,
        return_dto=user_read_dto,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def verify(token: str, service: UserServiceType, request: Request) -> SQLAUserT:
        """Verify a user with a given JWT."""

        return cast(SQLAUserT, await service.verify(token, request))

    return verify


def get_auth_handler(
    login_path: str,
    logout_path: str,
    user_read_dto: type[SQLAlchemyDTO],  # pyright: ignore
    auth_backend: JWTAuth | JWTCookieAuth | SessionAuth,
    authentication_schema: Any,
    tags: list[str] | None = None,
) -> Router:
    """Get authentication/login route handlers.

    Args:
        login_path: The path for the login router.
        logout_path: The path for the logout router.
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        auth_backend: A Litestar authentication backend.
        authentication_schema: The object that defines the request body schema.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(
        login_path,
        return_dto=user_read_dto,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def login_session(data: authentication_schema, service: UserServiceType, request: Request) -> SQLAUserT:
        """Authenticate a user."""
        if not isinstance(auth_backend, SessionAuth):
            raise ImproperlyConfiguredException("session login can only be used with SesssionAuth")

        user = await service.authenticate(data, request)
        if user is None:
            request.clear_session()
            raise NotAuthorizedException(detail="login failed, invalid input")

        request.set_session({"user_id": user.id})  # TODO: move and make configurable
        return cast(SQLAUserT, user)

    @post(
        login_path,
        return_dto=user_read_dto,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def login_jwt(data: authentication_schema, service: UserServiceType, request: Request) -> Response[SQLAUserT]:
        """Authenticate a user."""

        if not isinstance(auth_backend, (JWTAuth, JWTCookieAuth)):
            raise ImproperlyConfiguredException("jwt login can only be used with JWTAuth")

        user = await service.authenticate(data, request)
        if user is None:
            raise NotAuthorizedException(detail="login failed, invalid input")

        if user.is_verified is False:
            raise PermissionDeniedException(detail="not verified")

        return auth_backend.login(identifier=str(user.id), response_body=cast(SQLAUserT, user))

    @post(logout_path, tags=tags)
    async def logout(request: Request) -> None:
        """Log an authenticated user out."""
        request.clear_session()

    route_handlers = []
    if isinstance(auth_backend, SessionAuth):
        route_handlers.extend([login_session, logout])
    else:
        route_handlers.append(login_jwt)

    return Router(path="/", route_handlers=route_handlers)


def get_current_user_handler(
    path: str,
    user_read_dto: type[SQLAlchemyDTO],  # pyright: ignore
    user_update_dto: type[SQLAlchemyDTO],  # pyright: ignore
    tags: list[str] | None = None,
) -> Router:
    """Get current-user route handlers.

    Args:
        path: The path for the router.
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        user_update_dto: A subclass of [UserUpdateDTO][litestar_users.schema.UserUpdateDTO]
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @get(path, return_dto=user_read_dto, tags=tags)
    async def get_current_user(request: Request[SQLAUserT, Any, Any]) -> SQLAUserT:
        """Get current user info."""

        return request.user

    @patch(
        path,
        dto=user_update_dto,
        return_dto=user_read_dto,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def update_current_user(
        data: SQLAUserT,
        request: Request[SQLAUserT, Any, Any],
        service: UserServiceType,
    ) -> SQLAUserT:
        """Update the current user."""
        data.id = request.user.id  # type: ignore[assignment]
        return cast(SQLAUserT, await service.update_user(data=data))

    return Router(path="/", route_handlers=[get_current_user, update_current_user])


def get_password_reset_handler(forgot_path: str, reset_path: str, tags: list[str] | None = None) -> Router:
    """Get forgot-password and reset-password route handlers.

    Args:
        forgot_path: The path for the forgot-password router.
        reset_path: The path for the reset-password router.
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(
        forgot_path,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def forgot_password(data: ForgotPasswordSchema, service: UserServiceType) -> None:
        await service.initiate_password_reset(data.email)
        return

    @post(
        reset_path,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        exclude_from_auth=True,
        tags=tags,
    )
    async def reset_password(data: ResetPasswordSchema, service: UserServiceType) -> None:
        await service.reset_password(data.token, data.password)
        return

    return Router(path="/", route_handlers=[forgot_password, reset_password])


def get_user_management_handler(
    path_prefix: str,
    guards: list["Guard"],
    identifier_uri: str,
    opt: dict[str, Any],
    user_read_dto: type[SQLAlchemyDTO],  # pyright: ignore
    user_update_dto: type[SQLAlchemyDTO],  # pyright: ignore
    tags: list[str] | None = None,
) -> Router:
    """Get user management route handlers.

    Note:
        Routes are guarded by role authorization.

    Args:
        path_prefix: The path prefix for the routers.
        guards: List of Guard callables to determine who is authorized to manage users.
        identifier_uri: The path specifying the user ID and its type.
        opt: Optional route handler 'opts' to provide additional context to Guards.
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        user_update_dto: A subclass of [UserUpdateDTO][litestar_users.schema.UserUpdateDTO]
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @get(
        path=identifier_uri,
        dto=user_read_dto,
        return_dto=user_read_dto,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def get_user(user_id: Union[UUID, int], service: UserServiceType) -> SQLAUserT:
        """Get a user by id."""

        return cast(SQLAUserT, await service.get_user(user_id))

    @patch(
        path=identifier_uri,
        dto=user_update_dto,
        return_dto=user_read_dto,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def update_user(user_id: Union[UUID, int], data: SQLAUserT, service: UserServiceType) -> SQLAUserT:
        """Update a user's attributes."""
        data.id = user_id  # type: ignore[assignment]
        return cast(SQLAUserT, await service.update_user(data))

    @delete(
        path=identifier_uri,
        return_dto=user_read_dto,
        status_code=200,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def delete_user(user_id: Union[UUID, int], service: UserServiceType) -> SQLAUserT:
        """Delete a user from the database."""

        return cast(SQLAUserT, await service.delete_user(user_id))

    return Router(path=path_prefix, route_handlers=[get_user, update_user, delete_user])


def get_role_management_handler(
    path_prefix: str,
    assign_role_path: str,
    revoke_role_path: str,
    guards: list["Guard"],
    identifier_uri: str,
    opt: dict[str, Any],
    role_create_dto: type[SQLAlchemyDTO],  # pyright: ignore
    role_read_dto: type[SQLAlchemyDTO],  # pyright: ignore
    role_update_dto: type[SQLAlchemyDTO],  # pyright: ignore
    user_read_dto: type[SQLAlchemyDTO],  # pyright: ignore
    tags: list[str] | None = None,
) -> Router:
    """Get role management route handlers.

    Note:
        Routes are guarded by role authorization.

    Args:
        path_prefix: The path prefix for the routers.
        assign_role_path: The path for the role assignment router.
        revoke_role_path: The path for the role revokement router.
        guards: List of Guard callables to determine who is authorized to manage roles.
        identifier_uri: The path specifying the role ID and its type.
        opt: Optional route handler 'opts' to provide additional context to Guards.
        role_create_dto: A subclass of [RoleCreateDTO][litestar_users.schema.RoleCreateDTO]
        role_read_dto: A subclass of [RoleReadDTO][litestar_users.schema.RoleReadDTO]
        role_update_dto: A subclass of [RoleUpdateDTO][litestar_users.schema.RoleUpdateDTO]
        user_read_dto: A subclass of [UserReadDTO][litestar_users.schema.UserReadDTO]
        tags: A list of string tags to append to the schema of the route handlers.
    """

    @post(
        dto=role_create_dto,
        return_dto=role_read_dto,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def create_role(data: SQLARoleT, service: UserServiceType) -> SQLARoleT:
        """Create a new role."""
        return cast(SQLARoleT, await service.add_role(data))

    @patch(
        path=identifier_uri,
        dto=role_update_dto,
        return_dto=role_read_dto,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def update_role(role_id: Union[UUID, int], data: SQLARoleT, service: UserServiceType) -> SQLARoleT:
        """Update a role in the database."""
        data.id = role_id  # type: ignore[assignment]
        return cast(SQLARoleT, await service.update_role(role_id, data))

    @delete(
        path=identifier_uri,
        return_dto=role_read_dto,
        status_code=200,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def delete_role(role_id: Union[UUID, int], service: UserServiceType) -> SQLARoleT:
        """Delete a role from the database."""

        return cast(SQLARoleT, await service.delete_role(role_id))

    @put(
        return_dto=user_read_dto,
        path=assign_role_path,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def assign_role(data: UserRoleSchema, service: UserServiceType) -> SQLAUserT:
        """Assign a role to a user."""

        return cast(SQLAUserT, await service.assign_role(data.user_id, data.role_id))

    @put(
        return_dto=user_read_dto,
        path=revoke_role_path,
        guards=guards,
        opt=opt,
        dependencies={"service": Provide(provide_user_service, sync_to_thread=False)},
        tags=tags,
    )
    async def revoke_role(data: UserRoleSchema, service: UserServiceType) -> SQLAUserT:
        """Revoke a role from a user."""

        return cast(SQLAUserT, await service.revoke_role(data.user_id, data.role_id))

    return Router(path_prefix, route_handlers=[create_role, assign_role, revoke_role, update_role, delete_role])



from __future__ import annotations

from typing import Sequence, cast

from passlib.context import CryptContext

__all__ = ["PasswordManager"]


class PasswordManager:
    """Thin wrapper around `passlib`."""

    def __init__(self, hash_schemes: Sequence[str] | None = None) -> None:
        """Construct a PasswordManager.

        Args:
            hash_schemes: The encryption schemes to use. Defaults to ["argon2"].
        """
        if hash_schemes is None:
            hash_schemes = ["argon2"]
        self.context = CryptContext(schemes=hash_schemes, deprecated="auto")

    def hash(self, password: str) -> str:
        """Create a password hash.

        Args:
            password: The password to hash.
        """
        return cast("str", self.context.hash(password))

    def verify_and_update(self, password: str, password_hash: str | None) -> tuple[bool, str | None]:
        """Verify a password and rehash it if the hash is deprecated.

        Args:
            password: The password to verify.
            password_hash: The hash to verify against.
        """
        return cast("tuple[bool, str | None]", self.context.verify_and_update(password, password_hash))



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any
from uuid import UUID

from advanced_alchemy.exceptions import NotFoundError
from advanced_alchemy.extensions.litestar.plugins import SQLAlchemyAsyncConfig
from litestar.exceptions import ImproperlyConfiguredException
from sqlalchemy.orm import Load

from litestar_users.utils import get_litestar_users_plugin, get_sqlalchemy_plugin

__all__ = ["jwt_retrieve_user_handler", "session_retrieve_user_handler"]


if TYPE_CHECKING:
    from litestar.connection import ASGIConnection
    from litestar.contrib.jwt import Token
    from sqlalchemy import StatementLambdaElement

    from litestar_users.adapter.sqlalchemy.protocols import SQLAUserT
    from litestar_users.adapter.sqlalchemy.repository import SQLAlchemyUserRepository


def _get_user_repository(connection: ASGIConnection) -> SQLAlchemyUserRepository:
    sqlalchemy_config = get_sqlalchemy_plugin(connection.app)._config
    if not isinstance(sqlalchemy_config, SQLAlchemyAsyncConfig):
        raise ImproperlyConfiguredException("SQLAlchemy config must be of type `SQLAlchemyAsyncConfig`")
    async_session = sqlalchemy_config.provide_session(state=connection.app.state, scope=connection.scope)

    litestar_users_config = get_litestar_users_plugin(connection.app)._config
    return litestar_users_config.user_repository_class(
        session=async_session, model_type=litestar_users_config.user_model
    )


def _check_update_statement(connection: ASGIConnection, statement: StatementLambdaElement) -> StatementLambdaElement:
    if load_options := connection.route_handler.opt.get("user_load_options"):
        if not isinstance(load_options, Sequence):
            raise ValueError("user_load_options must be a sequence")
        if not all(isinstance(load_option, Load) for load_option in load_options):
            raise ValueError("all load options must be instances of `sqlalchemy.orm.Load`")
        return statement.options(*load_options)
    return statement


async def session_retrieve_user_handler(session: dict[str, Any], connection: ASGIConnection) -> SQLAUserT | None:
    """Get a user from the database based on session info.

    Args:
        session: Litestar session.
        connection: The ASGI connection.
    """
    repository = _get_user_repository(connection)
    try:
        # undocumented relationship loading api
        statement = _check_update_statement(connection, repository.statement)
        user_id = session.get("user_id")
        if user_id is None:
            return None
        user = await repository.get(UUID(user_id), statement=statement)
        if user.is_active and user.is_verified:
            return user  # type: ignore[no-any-return]
    except NotFoundError:
        pass
    return None


async def jwt_retrieve_user_handler(token: Token, connection: ASGIConnection) -> SQLAUserT | None:
    """Get a user from the database based on JWT info.

    Args:
        token: Encoded JWT.
        connection: The ASGI connection.
    """
    repository = _get_user_repository(connection)
    try:
        # undocumented relationship loading api
        statement = _check_update_statement(connection, repository.statement)
        user_id: UUID | int | None = None
        try:
            user_id = UUID(token.sub)
        except ValueError:
            user_id = int(token.sub)
        user = await repository.get(user_id, statement=statement)
        if user.is_active and user.is_verified:
            return user  # type: ignore[no-any-return]
    except NotFoundError:
        pass
    return None



from dataclasses import dataclass
from uuid import UUID

__all__ = [
    "ForgotPasswordSchema",
    "ResetPasswordSchema",
    "AuthenticationSchema",
    "UserRoleSchema",
]


@dataclass
class AuthenticationSchema:
    """User authentication schema."""

    email: str
    password: str


@dataclass
class ForgotPasswordSchema:
    """Forgot password schema."""

    email: str


@dataclass
class ResetPasswordSchema:
    """Reset password schema."""

    token: str
    password: str


@dataclass
class UserRoleSchema:
    """User role association schema."""

    user_id: UUID
    role_id: UUID



from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Generic, Sequence, TypeVar
from uuid import UUID

from advanced_alchemy.exceptions import ConflictError, NotFoundError
from jose import JWTError
from litestar.contrib.jwt.jwt_token import Token
from litestar.exceptions import ImproperlyConfiguredException

from litestar_users.adapter.sqlalchemy.protocols import SQLARoleT, SQLAUserT
from litestar_users.exceptions import InvalidTokenException
from litestar_users.password import PasswordManager

__all__ = ["BaseUserService"]


if TYPE_CHECKING:
    from litestar import Request

    from litestar_users.adapter.sqlalchemy.repository import SQLAlchemyRoleRepository, SQLAlchemyUserRepository
    from litestar_users.schema import AuthenticationSchema


class BaseUserService(Generic[SQLAUserT, SQLARoleT]):  # pylint: disable=R0904
    """Main user management interface."""

    user_model: type[SQLAUserT]
    """A subclass of the `User` ORM model."""

    def __init__(
        self,
        secret: str,
        user_auth_identifier: str,
        user_repository: SQLAlchemyUserRepository[SQLAUserT],
        hash_schemes: Sequence[str] | None = None,
        role_repository: SQLAlchemyRoleRepository[SQLARoleT, SQLAUserT] | None = None,
    ) -> None:
        """User service constructor.

        Args:
            secret: Secret string for securely signing tokens.
            user_auth_identifier: The `User` model attribute to identify the user during authorization.
            user_repository: A `UserRepository` instance.
            hash_schemes: Schemes to use for password encryption.
            role_repository: A `RoleRepository` instance.
        """
        self.user_repository = user_repository
        self.role_repository = role_repository
        self.secret = secret
        self.password_manager = PasswordManager(hash_schemes=hash_schemes)
        self.user_model = self.user_repository.model_type
        if role_repository is not None:
            self.role_model = role_repository.model_type
        self.user_auth_identifier = user_auth_identifier

    async def add_user(self, user: SQLAUserT, verify: bool = False, activate: bool = True) -> SQLAUserT:
        """Create a new user programmatically.

        Args:
            user: User model instance.
            verify: Set the user's verification status to this value.
            activate: Set the user's active status to this value.
        """
        existing_user = await self.get_user_by(email=user.email)
        if existing_user:
            raise ConflictError("email already associated with an account")

        user.is_verified = verify
        user.is_active = activate

        return await self.user_repository.add(user)

    async def register(self, data: dict[str, Any], request: Request | None = None) -> SQLAUserT:
        """Register a new user and optionally run custom business logic.

        Args:
            data: User creation data transfer object.
            request: The litestar request that initiated the action.
        """
        await self.pre_registration_hook(data, request)

        data["password_hash"] = self.password_manager.hash(data.pop("password"))
        user = await self.add_user(self.user_model(**data))  # type: ignore[arg-type]
        await self.initiate_verification(user)  # TODO: make verification optional?

        await self.post_registration_hook(user, request)

        return user

    async def get_user(self, id_: UUID | int) -> SQLAUserT:
        """Retrieve a user from the database by id.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        return await self.user_repository.get(id_)

    async def get_user_by(self, **kwargs: Any) -> SQLAUserT | None:
        """Retrieve a user from the database by arbitrary keyword arguments.

        Args:
            **kwargs: Keyword arguments to pass as filters.

        Examples:
            ```python
            service = UserService(...)
            john = await service.get_one(email="john@example.com")
            ```
        """
        return await self.user_repository.get_one_or_none(**kwargs)

    async def update_user(self, data: SQLAUserT) -> SQLAUserT:
        """Update arbitrary user attributes in the database.

        Args:
            data: User update data transfer object.
        """

        return await self.user_repository.update(data)

    async def delete_user(self, id_: UUID | int) -> SQLAUserT:
        """Delete a user from the database.

        Args:
            id_: UUID corresponding to a user primary key.
        """
        return await self.user_repository.delete(id_)

    async def authenticate(self, data: Any, request: Request | None = None) -> SQLAUserT | None:
        """Authenticate a user.

        Args:
            data: User authentication data transfer object.
            request: The litestar request that initiated the action.
        """
        # avoid early returns to mitigate timing attacks.
        # check if user supplied logic should allow authentication, but only
        # supply the result later.
        should_proceed = await self.pre_login_hook(data, request)

        try:
            user = await self.user_repository.get_one(
                **{self.user_auth_identifier: getattr(data, self.user_auth_identifier)}
            )
        except NotFoundError:
            # trigger passlib's `dummy_verify` method
            self.password_manager.verify_and_update(data.password, None)
            return None

        password_verified, new_password_hash = self.password_manager.verify_and_update(
            data.password, user.password_hash
        )
        if new_password_hash is not None:
            user = await self.user_repository._update(user, {"password_hash": new_password_hash})

        if not password_verified or not should_proceed:
            return None

        await self.post_login_hook(user, request)

        return user

    def generate_token(self, user_id: UUID | int, aud: str) -> str:
        """Generate a limited time valid JWT.

        Args:
            user_id: ID of the user to provide the token to.
            aud: Context of the token
        """
        token = Token(
            exp=datetime.now() + timedelta(seconds=60 * 60 * 24),  # noqa: DTZ005
            sub=str(user_id),
            aud=aud,
        )
        return token.encode(secret=self.secret, algorithm="HS256")

    async def initiate_verification(self, user: SQLAUserT) -> None:
        """Initiate the user verification flow.

        Args:
            user: The user requesting verification.
        """
        token = self.generate_token(user.id, aud="verify")
        await self.send_verification_token(user, token)

    async def send_verification_token(self, user: SQLAUserT, token: str) -> None:
        """Execute custom logic to send the verification token to the relevant user.

        Args:
            user: The user requesting verification.
            token: An encoded JWT bound to verification.

        Notes:
        - Develepors need to override this method to facilitate sending the token via email, sms etc.
        """

    async def verify(self, encoded_token: str, request: Request | None = None) -> SQLAUserT:
        """Verify a user with the given JWT.

        Args:
            encoded_token: An encoded JWT bound to verification.
            request: The litestar request that initiated the action.

        Raises:
            InvalidTokenException: If the token is expired or tampered with.
        """
        token = self._decode_and_verify_token(encoded_token, context="verify")

        try:
            user_id: UUID | int = UUID(token.sub)
        except ValueError:
            user_id = int(token.sub)
        try:
            user = await self.user_repository.update(self.user_model(id=user_id, is_verified=True))  # type: ignore[arg-type]
        except NotFoundError as e:
            raise InvalidTokenException("token is invalid") from e

        await self.post_verification_hook(user, request)

        return user

    async def initiate_password_reset(self, email: str) -> None:
        """Initiate the password reset flow.

        Args:
            email: Email of the user who has forgotten their password.
        """
        user = await self.get_user_by(email=email)  # TODO: something about timing attacks.
        if user is None:
            return
        token = self.generate_token(user.id, aud="reset_password")
        await self.send_password_reset_token(user, token)

    async def send_password_reset_token(self, user: SQLAUserT, token: str) -> None:
        """Execute custom logic to send the password reset token to the relevant user.

        Args:
            user: The user requesting the password reset.
            token: An encoded JWT bound to the password reset flow.

        Notes:
        - Develepors need to override this method to facilitate sending the token via email, sms etc.
        """

    async def reset_password(self, encoded_token: str, password: str) -> None:
        """Reset a user's password given a valid JWT.

        Args:
            encoded_token: An encoded JWT bound to the password reset flow.
            password: The new password to hash and store.

        Raises:
            InvalidTokenException: If the token has expired or been tampered with.
        """
        token = self._decode_and_verify_token(encoded_token, context="reset_password")

        try:
            user_id: UUID | int = UUID(token.sub)
        except ValueError:
            user_id = int(token.sub)
        try:
            await self.user_repository.update(
                self.user_model(id=user_id, password_hash=self.password_manager.hash(password))  # type: ignore[arg-type]
            )
        except NotFoundError as e:
            raise InvalidTokenException from e

    async def pre_login_hook(
        self, data: AuthenticationSchema, request: Request | None = None
    ) -> bool:  # pylint: disable=W0613
        """Execute custom logic to run custom business logic prior to authenticating a user.

        Useful for authentication checks against external sources,
        eg. current membership validity or blacklists, etc
        Must return `False` or raise a custom exception to cancel authentication.

        Args:
            data: Authentication data transfer object.
            request: The litestar request that initiated the action.

        Notes:
            Uncaught exceptions in this method will break the authentication process.
        """
        return True

    async def post_login_hook(self, user: SQLAUserT, request: Request | None = None) -> None:
        """Execute custom logic to run custom business logic after authenticating a user.

        Useful for eg. updating a login counter, updating last known user IP
        address, etc.

        Args:
            user: The user who has authenticated.
            request: The litestar request that initiated the action.

        Notes:
            Uncaught exceptions in this method will break the authentication process.
        """
        return

    async def pre_registration_hook(
        self, data: dict[str, Any], request: Request | None = None
    ) -> None:  # pylint: disable=W0613
        """Execute custom logic to run custom business logic prior to registering a user.

        Useful for authorization checks against external sources,
        eg. membership API or blacklists, etc.

        Args:
            data: User creation data transfer object
            request: The litestar request that initiated the action.

        Notes:
        - Uncaught exceptions in this method will result in failed registration attempts.
        """
        return

    async def post_registration_hook(self, user: SQLAUserT, request: Request | None = None) -> None:
        """Execute custom logic to run custom business logic after registering a user.

        Useful for updating external datasets, sending welcome messages etc.

        Args:
            user: User ORM instance.
            request: The litestar request that initiated the action.

        Notes:
        - Uncaught exceptions in this method could result in returning a HTTP 500 status
        code while successfully creating the user in the database.
        - It's possible to skip verification entirely by setting `user.is_verified`
        to `True` here.
        """
        return

    async def post_verification_hook(self, user: SQLAUserT, request: Request | None = None) -> None:
        """Execute custom logic to run custom business logic after a user has verified details.

        Useful for eg. updating sales lead data, etc.

        Args:
            user: User ORM instance.
            request: The litestar request that initiated the action.

        Notes:
        - Uncaught exceptions in this method could result in returning a HTTP 500 status
        code while successfully validating the user.
        """
        return

    def _decode_and_verify_token(self, encoded_token: str, context: str) -> Token:
        try:
            token = Token.decode(
                encoded_token=encoded_token,
                secret=self.secret,
                algorithm="HS256",
            )
        except JWTError as e:
            raise InvalidTokenException from e

        if token.aud != context:
            raise InvalidTokenException(f"aud value must be {context}")

        return token

    async def get_role(self, id_: UUID | int) -> SQLARoleT:
        """Retrieve a role by id.

        Args:
            id_: ID of the role.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        return await self.role_repository.get(id_)

    async def get_role_by_name(self, name: str) -> SQLARoleT:
        """Retrieve a role by name.

        Args:
            name: The name of the role.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        return await self.role_repository.get_one(name=name)

    async def add_role(self, data: SQLARoleT) -> SQLARoleT:
        """Add a new role to the database.

        Args:
            data: A role creation data transfer object.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        return await self.role_repository.add(data)

    async def update_role(self, id_: UUID | int, data: SQLARoleT) -> SQLARoleT:
        """Update a role in the database.

        Args:
            id_: UUID corresponding to the role primary key.
            data: A role update data transfer object.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        return await self.role_repository.update(data)

    async def delete_role(self, id_: UUID | int) -> SQLARoleT:
        """Delete a role from the database.

        Args:
            id_: UUID corresponding to the role primary key.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        return await self.role_repository.delete(id_)

    async def assign_role(self, user_id: UUID | int, role_id: UUID | int) -> SQLAUserT:
        """Add a role to a user.

        Args:
            user_id: ID of the user to receive the role.
            role_id: ID of the role to add to the user.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        user = await self.get_user(user_id)
        role = await self.get_role(role_id)

        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("roles have not been configured")

        if isinstance(user.roles, list) and role in user.roles:  # pyright: ignore
            raise ConflictError(f"user already has role '{role.name}'")
        return await self.role_repository.assign_role(user, role)

    async def revoke_role(self, user_id: UUID | int, role_id: UUID | int) -> SQLAUserT:
        """Revoke a role from a user.

        Args:
            user_id: ID of the user to revoke the role from.
            role_id: ID of the role to revoke.
        """
        if self.role_repository is None:
            raise ImproperlyConfiguredException("roles have not been configured")
        user = await self.get_user(user_id)
        role = await self.get_role(role_id)

        if not hasattr(user, "roles"):
            raise ImproperlyConfiguredException("roles have not been configured")

        if isinstance(user.roles, list) and role not in user.roles:  # pyright: ignore
            raise ConflictError(f"user does not have role '{role.name}'")
        return await self.role_repository.revoke_role(user, role)


UserServiceType = TypeVar("UserServiceType", bound=BaseUserService)



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence
from uuid import UUID

from advanced_alchemy.exceptions import RepositoryError
from advanced_alchemy.types import GUID
from litestar.contrib.jwt import JWTAuth, JWTCookieAuth
from litestar.dto import DTOData
from litestar.plugins import CLIPluginProtocol, InitPluginProtocol
from litestar.security.session_auth import SessionAuth
from sqlalchemy.sql.sqltypes import BigInteger, Uuid

from litestar_users.exceptions import TokenException, repository_exception_to_http_response, token_exception_handler
from litestar_users.route_handlers import (
    get_auth_handler,
    get_current_user_handler,
    get_password_reset_handler,
    get_registration_handler,
    get_role_management_handler,
    get_user_management_handler,
    get_verification_handler,
)
from litestar_users.schema import ForgotPasswordSchema, ResetPasswordSchema, UserRoleSchema
from litestar_users.user_handlers import (
    jwt_retrieve_user_handler,
    session_retrieve_user_handler,
)

__all__ = ["LitestarUsersPlugin"]


if TYPE_CHECKING:
    from click import Group
    from litestar import Router
    from litestar.config.app import AppConfig
    from litestar.handlers import HTTPRouteHandler
    from litestar.types import ExceptionHandlersMap

    from litestar_users.config import LitestarUsersConfig


class LitestarUsersPlugin(InitPluginProtocol, CLIPluginProtocol):
    """A Litestar extension for authentication, authorization and user management."""

    def __init__(self, config: LitestarUsersConfig) -> None:
        """Construct a LitestarUsers instance."""
        self._config = config

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Register routers, auth strategies etc on the Litestar app.

        Args:
            app_config: An instance of [AppConfig][litestar.config.AppConfig]
        """
        auth_backend = self._get_auth_backend()
        route_handlers = self._get_route_handlers(auth_backend)

        app_config = auth_backend.on_app_init(app_config)
        app_config.route_handlers.extend(route_handlers)

        exception_handlers: ExceptionHandlersMap = {
            RepositoryError: repository_exception_to_http_response,
            TokenException: token_exception_handler,
        }
        app_config.exception_handlers.update(exception_handlers)

        app_config.signature_namespace.update(
            {
                "ForgotPasswordSchema": ForgotPasswordSchema,
                "ResetPasswordSchema": ResetPasswordSchema,
                "authentication_schema": self._config.authentication_request_schema,
                "UserRoleSchema": UserRoleSchema,
                "UserServiceType": self._config.user_service_class,
                "BaseUserService": self._config.user_service_class,
                "SQLAUserT": self._config.user_model,
                "SQLARoleT": self._config.role_model,
                "role_create_dto": self._config.role_create_dto,
                "role_read_dto": self._config.role_read_dto,
                "role_update_dto": self._config.role_update_dto,
                "user_read_dto": self._config.user_read_dto,
                "user_update_dto": self._config.user_update_dto,
                "user_registration_dto": self._config.user_registration_dto,
                "DTOData": DTOData,
                "UserRegisterT": self._config.user_registration_dto.model_type,  # type: ignore[misc]
                "UUID": UUID,
            }
        )
        app_config.state.update({"litestar_users_config": self._config})

        return app_config

    def on_cli_init(self, cli: Group) -> None:
        from litestar_users.cli import user_management_group

        cli.add_command(user_management_group)
        return super().on_cli_init(cli)

    def get_user_identifier_uri(self) -> str:
        if isinstance(self._config.user_model.id.type, (GUID, Uuid)):
            return "/{user_id:uuid}"
        if isinstance(self._config.user_model.id.type, BigInteger):
            return "/{user_id:int}"
        raise ValueError("user identifier type not supported")

    def get_role_identifier_uri(self) -> str:
        if self._config.role_model is None:
            return ""
        if isinstance(self._config.role_model.id.type, (GUID, Uuid)):
            return "/{role_id:uuid}"
        if isinstance(self._config.role_model.id.type, BigInteger):
            return "/{role_id:int}"
        raise ValueError("role identifier type not supported")

    def _get_auth_backend(self) -> JWTAuth | JWTCookieAuth | SessionAuth:
        if issubclass(self._config.auth_backend_class, SessionAuth):
            return self._config.auth_backend_class(
                retrieve_user_handler=session_retrieve_user_handler,
                session_backend_config=self._config.session_backend_config,  # type: ignore
                exclude=self._config.auth_exclude_paths,
            )
        if issubclass(self._config.auth_backend_class, JWTAuth) or issubclass(
            self._config.auth_backend_class, JWTCookieAuth
        ):
            return self._config.auth_backend_class(
                retrieve_user_handler=jwt_retrieve_user_handler,
                token_secret=self._config.secret,
                exclude=self._config.auth_exclude_paths,
            )
        raise ValueError("invalid auth backend")

    def _get_route_handlers(
        self, auth_backend: JWTAuth | JWTCookieAuth | SessionAuth
    ) -> Sequence[HTTPRouteHandler | Router]:
        """Parse the route handler configs to get Routers."""

        handlers: list[HTTPRouteHandler | Router] = []
        if self._config.auth_handler_config:
            handlers.append(
                get_auth_handler(
                    login_path=self._config.auth_handler_config.login_path,
                    logout_path=self._config.auth_handler_config.logout_path,
                    user_read_dto=self._config.user_read_dto,
                    auth_backend=auth_backend,
                    authentication_schema=self._config.authentication_request_schema,
                    tags=self._config.auth_handler_config.tags,
                )
            )
        if self._config.current_user_handler_config:
            handlers.append(
                get_current_user_handler(
                    path=self._config.current_user_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    tags=self._config.current_user_handler_config.tags,
                )
            )
        if self._config.password_reset_handler_config:
            handlers.append(
                get_password_reset_handler(
                    forgot_path=self._config.password_reset_handler_config.forgot_path,
                    reset_path=self._config.password_reset_handler_config.reset_path,
                    tags=self._config.password_reset_handler_config.tags,
                )
            )
        if self._config.register_handler_config:
            handlers.append(
                get_registration_handler(
                    path=self._config.register_handler_config.path,
                    user_registration_dto=self._config.user_registration_dto,
                    user_read_dto=self._config.user_read_dto,
                    tags=self._config.register_handler_config.tags,
                )
            )
        if self._config.role_management_handler_config:
            handlers.append(
                get_role_management_handler(
                    path_prefix=self._config.role_management_handler_config.path_prefix,
                    assign_role_path=self._config.role_management_handler_config.assign_role_path,
                    revoke_role_path=self._config.role_management_handler_config.revoke_role_path,
                    guards=self._config.role_management_handler_config.guards,
                    identifier_uri=self.get_role_identifier_uri(),
                    opt=self._config.role_management_handler_config.opt,
                    role_create_dto=self._config.role_create_dto,  # type: ignore[arg-type]
                    role_read_dto=self._config.role_read_dto,  # type: ignore[arg-type]
                    role_update_dto=self._config.role_update_dto,  # type: ignore[arg-type]
                    user_read_dto=self._config.user_read_dto,
                    tags=self._config.role_management_handler_config.tags,
                )
            )
        if self._config.user_management_handler_config:
            handlers.append(
                get_user_management_handler(
                    path_prefix=self._config.user_management_handler_config.path_prefix,
                    guards=self._config.user_management_handler_config.guards,
                    identifier_uri=self.get_user_identifier_uri(),
                    opt=self._config.user_management_handler_config.opt,
                    user_read_dto=self._config.user_read_dto,
                    user_update_dto=self._config.user_update_dto,
                    tags=self._config.user_management_handler_config.tags,
                )
            )
        if self._config.verification_handler_config:
            handlers.append(
                get_verification_handler(
                    path=self._config.verification_handler_config.path,
                    user_read_dto=self._config.user_read_dto,
                    tags=self._config.verification_handler_config.tags,
                )
            )
        return handlers



from __future__ import annotations

from typing import TYPE_CHECKING

from advanced_alchemy.extensions.litestar.plugins import SQLAlchemyAsyncConfig
from litestar.exceptions import ImproperlyConfiguredException

from litestar_users.adapter.sqlalchemy.repository import SQLAlchemyRoleRepository
from litestar_users.utils import get_litestar_users_plugin, get_sqlalchemy_plugin

__all__ = ["provide_user_service"]


if TYPE_CHECKING:
    from litestar import Request
    from litestar.datastructures import State

    from litestar_users.service import BaseUserService


def provide_user_service(state: State, request: Request) -> BaseUserService:
    """Instantiate service and repository for use with DI.

    Args:
        request: The incoming request
        state: The application.state instance
    """

    sqlalchemy_config = get_sqlalchemy_plugin(request.app)._config
    if not isinstance(sqlalchemy_config, SQLAlchemyAsyncConfig):
        raise ImproperlyConfiguredException("SQLAlchemy config must be of type `SQLAlchemyAsyncConfig`")
    session = sqlalchemy_config.provide_session(state=state, scope=request.scope)

    litestar_users_config = get_litestar_users_plugin(request.app)._config
    user_repository = litestar_users_config.user_repository_class(
        session=session, model_type=litestar_users_config.user_model
    )
    role_repository: SQLAlchemyRoleRepository | None = (
        None
        if litestar_users_config.role_model is None
        else SQLAlchemyRoleRepository(session=session, model_type=litestar_users_config.role_model)
    )

    return litestar_users_config.user_service_class(
        user_auth_identifier=litestar_users_config.user_auth_identifier,
        user_repository=user_repository,
        role_repository=role_repository,
        secret=litestar_users_config.secret,
        hash_schemes=litestar_users_config.hash_schemes,
    )
